# -*- coding: UTF-8 -*-

import nes_common.logger as logger


class _Insert(object):
    cache = {}
    cacheValues = {}
    cacheOddColumns = {}  # кэш со столбцами, которые пропускаем (не пытаемся грузить в БД)
    cacheOwnerStoreIdPresented = {}  # если поле ownerStoreId предоставлено в таблице, при insert его нужно установить в NULL, в ином случае при обогащении данных
                               # из All запись для магазина не будет помечена, что она теперь из магазина и ее нужно выгружать

    def __init__(self, table, values, quotedValues=None, idName=None, replaceSymbol='%s', saveLastId=False, onDuplicateKeyUpdate=False):
        self.table = table
        self.values = values
        self.quotedValues = quotedValues
        self.id = None
        self.replaceSymbol = replaceSymbol
        self.saveLastId = saveLastId
        self.replace = False
        self.cacheKey = table
        self.onDuplicateKeyUpdate = onDuplicateKeyUpdate

    def __getitem__(self, x):
        return self.values[x]

    def __setitem__(self, x, v):
        self.values[x] = v

    def __str__(self):
        return '[INSERT table=%s, values=%s, quotedValues=%s]' % (self.table, self.values, self.quotedValues)

    def prepareStatement(self, context):
        if self.cacheKey in self.cache:
            sql = _Insert.cache[self.cacheKey]
        else:
            sqlCommand = 'replace' if self.replace else 'insert ignore' if context.executeMany else 'insert'
            tableName = context.tableName(self.table)
            fields = self.values.keys()
            fieldsCount = len(fields)
            if self.quotedValues is not None:
                quotedFields = self.quotedValues.keys()
                sql = '%s into %s (`%s`, `%s`) values (%s, %s)' % (sqlCommand, tableName, '`,`'.join(fields),
                        ','.join(quotedFields), '`,`'.join([self.replaceSymbol]*fieldsCount),
                        ','.join(self.quotedValues.values()))
            else:
                sql = '%s into %s (`%s`) values (%s)' % (sqlCommand, tableName, '`,`'.join(fields),
                        ','.join([self.replaceSymbol]*fieldsCount))
            if self.onDuplicateKeyUpdate:
                sql += (' on duplicate key update %s' % (self.getOnDuplicateKeyClause(fields)))
            _Insert.cache[self.cacheKey] = sql
            _Insert.cacheValues[self.cacheKey] = []
        return sql

    def execute(self, context):
        values = self.values.values()
        sql = self.prepareStatement(context)
        if context.executeMany and not self.saveLastId:
            _Insert.cacheValues[self.cacheKey].append([v() if callable(v) else v for v in values])
        else:
            context.cursor.execute(sql, [v() if callable(v) else v for v in values])
            if self.saveLastId:
                context.cursor.execute('SELECT LAST_INSERT_ID()')
                for value in context.cursor:
                    self.id = value[0]

    def execMany(self, context):
        for table in _Insert.cache.keys():
            context.cursor.executemany(_Insert.cache[table], _Insert.cacheValues[table])
            del _Insert.cacheValues[table][:]

    def lastId(self):
        return self.id

    def filterColumns(self, generateMap, context):
        if self.table not in _Insert.cacheOddColumns:
            _Insert.cacheOddColumns[self.table] = []
            realColumns = self.getRealColumns(context)
            _Insert.cacheOwnerStoreIdPresented[self.table] = "ownerstoreid" in realColumns
            for k in generateMap.keys():
                if not k.lower() in realColumns:
                    _Insert.cacheOddColumns[self.table].append(k)
                    generateMap.pop(k)
            if len(_Insert.cacheOddColumns[self.table]) > 0:
                log = logger.getLogger('Shop', context.shopCode if hasattr(context, 'shopCode') else 'cash')
                log.warn('Не загружаем поля: %s в таблицу: %s' % (_Insert.cacheOddColumns[self.table], self.table))
        else:
            for k in _Insert.cacheOddColumns[self.table]:
                generateMap.pop(k)

    def getRealColumns(self, context):
        columns = set()
        context.cursor.execute('SHOW COLUMNS FROM %s' % self.table)
        for value in context.cursor:
            columns.add(value[0].lower())
        return columns

    def getOnDuplicateKeyClause(self, fields):
        onDuplicateKeyClause = ""
        for field in fields:
            if onDuplicateKeyClause:
                onDuplicateKeyClause += ', '
            onDuplicateKeyClause += '`{0}` = values(`{0}`)'.format(field)
        return onDuplicateKeyClause


class _Update(object):
    cache = {}

    def __init__(self, table, fields, values, whereField, whereValue, replaceSymbol='%s'):
        self.table = table        
        self.fields = fields
        self.values = values
        self.whereField = whereField
        self.whereValue = whereValue
        self.replaceSymbol = replaceSymbol
        
    def __str__(self):
        return '[UPDATE table=%s, fields=%s, values=%s, whereField=%s, whereValue=%s]' % (self.table, self.fields, self.values, self.whereField, self.whereValue)
    
    def prepareStatement(self, context):
        fields = self.fields
        values = self.values
        whereField = self.whereField
        whereValue = self.whereValue
        tableName = context.tableName(self.table)
        setClause = (', '.join(["%s = '?'"] * len(fields)) % tuple(fields)).replace('?', '%s')
        setClause = setClause % tuple(values)
        sql = ('update %s set %s where %s in (?)' % (tableName, setClause, whereField)).replace('?', self.replaceSymbol)
        return sql
    
    def execute(self, context):
        sql = self.prepareStatement(context)
        if isinstance(self.whereValue, _Select):
            whereValue = map(lambda x: list(x), self.whereValue.execute(context))
            context.cursor.executemany(sql, whereValue)
        else:
            whereValue = list(self.whereValue)
            context.cursor.execute(sql, whereValue)


class _LastInsertId(object):
    value = None
    
    def __init__(self):
        self.value = None

    def __str__(self):
        return '[SELECT LAST_INSERT_ID()]'

    def execute(self, context):
        if self.value is not None:
            return self.value
        sql = 'SELECT LAST_INSERT_ID()'
        context.cursor.execute(sql)
        for value in context.cursor:
            self.value = value[0]
        return self.value

    def id(self):
        return lambda: self.value


class _Select(object):
    result = False
    cache = {}

    def __init__(self, table, fields, whereFields, whereValues, replaceSymbol='%s', arbitraryCondition=''):
        self.table = table
        self.fields = fields
        self.whereFields = whereFields
        self.whereValues = whereValues
        self.result = False
        self.replaceSymbol = replaceSymbol
        self.arbitraryCondition = arbitraryCondition

    def __str__(self):
        return '[SELECT table=%s, fields=%s, whereFields=%s, whereValues=%s]' % (self.table, self.fields, self.whereFields, self.whereValues)
    
    def prepare(self, context):
        if self.table in _Select.cache:
            return _Select.cache[self.table]
        tableName = context.tableName(self.table)
        whereFields = self.whereFields  
        whereClause = (' and '.join(['%s = ?'] * len(whereFields)) % tuple(whereFields)).replace('?', self.replaceSymbol)
        sql = 'select %s from %s where %s %s' % (','.join(self.fields), tableName, whereClause, self.arbitraryCondition)
        _Select.cache[self.table] = sql      
        return sql
            
    def execute(self, context):
        if self.result:
            return self.result
        sql = self.prepare(context)
        if isinstance(self.whereValues, _Select):
            whereValues = map(lambda x: list(x), self.whereValues.execute(context))
            result = []
            for val in whereValues:
                context.cursor.execute(sql, val)
                result += context.cursor.fetchall()
        else:
            whereValues = [str(val.encode("utf8") if isinstance(val, unicode) else val) for val in list(self.whereValues)]
            context.cursor.execute(sql, whereValues)
            result = context.cursor.fetchall()
        self.result = result        
        return result
    
    def oneValue(self):
        if self.result is not None:
            return self.result[0][0]
        else:
            return None


class _Delete(object):
    cache = {}
    cacheValues = {}

    def __init__(self, table, whereFields, whereValues, replaceSymbol='%s', notExecMany=False):
        self.table = table
        self.whereFields = whereFields
        self.whereValues = whereValues
        self.replaceSymbol = replaceSymbol
        self.notExecMany = notExecMany

    def __str__(self):
        return '[DELETE table=%s, whereFields=%s, whereValues=%s]' % (self.table, self.whereFields, self.whereValues)

    def prepareStatement(self, context):
        hashkey = self.table + '@' + repr(self.whereFields)
        if hashkey in _Delete.cache:
            sql = _Delete.cache[hashkey]
        else:
            tableName = context.tableName(self.table)
            whereFields = self.whereFields
            whereClause = (' and '.join(['%s = ?'] * len(whereFields)) % tuple(whereFields)).replace('?', self.replaceSymbol)
            sql = 'delete from %s where %s' % (tableName, whereClause)
            _Delete.cache[hashkey] = sql
            _Delete.cacheValues[hashkey] = []
        return sql, hashkey
    
    def execute(self, context):
        if self.table in context.emptyTables:
            return
        sql, hashkey = self.prepareStatement(context)
#        context.cursor.execute('FLUSH tables;')
        if isinstance(self.whereValues, _Select):
            whereValues = map(lambda x: list(x), self.whereValues.execute(context))
            if len(whereValues) > 0:
                if context.executeMany and not self.notExecMany:
                    _Delete.cacheValues[hashkey].extend(whereValues)
                else:
                    context.cursor.executemany(sql, whereValues)
        else:
            whereValues = []
            for val in list(self.whereValues):  
                whereValues.append([str(value) for value in val.values()] if isinstance(val, dict) else str(val))
            if context.executeMany and not self.notExecMany:
                _Delete.cacheValues[hashkey].append(whereValues)
            else:
                context.cursor.execute(sql, whereValues)

    def execMany(self, context):
        for key in _Delete.cache.keys():
            context.cursor.executemany(_Delete.cache[key], _Delete.cacheValues[key])
            del _Delete.cacheValues[key][:]


class _DeleteAll(object):
    def __init__(self, table):
        self.table = table

    def __str__(self):
        return "[DELETEALL table=%s]" % (self.table,)
    
    def execute(self, context):
        tableName = context.tableName(self.table)
        sql = "truncate %s " % tableName
        context.cursor.execute(sql)
        context.emptyTables[self.table] = True
    
    def execMany(self, context):
        self.execute(context)


class _DeleteByDate(_Delete):
    def prepareStatement(self, context):
        self.notExecMany = True
        # если дата окончания переоценки равна None, то ключ должен быть другим
        hashkey = self.table + '@' + repr(self.whereFields if self.whereValues[2] else self.whereFields[0:2]) 
        if hashkey in _Delete.cache:
            sql = _Delete.cache[hashkey]
        else:
            tableName = context.tableName(self.table)
            # если дата окончания переоценки равна None, то условие следует изменить
            whereClause = (' barcode = ? and effectivedate < ? and effectivedateend %s' % ('= ?' if self.whereValues[2] else 'is null')).replace('?', self.replaceSymbol)
            sql = 'delete from %s where %s' % (tableName, whereClause)
        _Delete.cache[hashkey] = sql
        _Delete.cacheValues[hashkey] = []
        return sql, hashkey

    def execute(self, context):
        if self.table in context.emptyTables:
            return
        sql, hashkey = self.prepareStatement(context)
        whereValues = [str(val.encode("utf8") if isinstance(val, unicode) else val) for val in list(self.whereValues) if val]
        if context.executeMany and not self.notExecMany:
            _Delete.cacheValues[hashkey].append(whereValues)
        else:
            context.cursor.execute(sql, whereValues)


class _DeleteOverduePrices(_Delete):
    def prepareStatement(self, context):
        hashkey = self.table + '@' + repr(self.whereFields)
        if hashkey in _Delete.cache:
            sql = _Delete.cache[hashkey]
        else:
            tableName = context.tableName(self.table)
            sql = ('delete from %s where barcode = ? and effectivedateend < now()' % tableName).replace('?', self.replaceSymbol)
            _Delete.cache[hashkey] = sql
            _Delete.cacheValues[hashkey] = []
        return sql, hashkey


class _DeleteBeforeDate(_Delete):
    def __init__(self, table, whereFields, whereValues, dateField, dateValue=None, replaceSymbol='%s', notExecMany=False):
        _Delete.__init__(self, table, whereFields, whereValues, replaceSymbol, notExecMany)
        self.dateField = dateField
        self.dateValue = dateValue

    def __str__(self):
        return '[DELETE table=%s, whereFields=%s, whereValues=%s, dateField=%s before dateValue=%s]' % (self.table, self.whereFields, self.whereValues, self.dateField, self.dateValue)

    def prepareStatement(self, context):
        hashkey = self.table + '@' + repr(self.whereFields)+'before'+repr(self.dateField)
        if self.dateValue:
            self.whereValues.append(self.dateValue)
        if hashkey in _Delete.cache:
            sql = _Delete.cache[hashkey]
        else:
            tableName = context.tableName(self.table)
            whereFields = self.whereFields
            whereClause = (' and '.join(['%s = ?'] * len(whereFields)) % tuple(whereFields)) + (' AND %s < ?' % self.dateField)
            whereClause = whereClause.replace('?', self.replaceSymbol)
            sql = 'delete from %s where %s' % (tableName, whereClause)
            _Delete.cache[hashkey] = sql
            _Delete.cacheValues[hashkey] = []
        return sql, hashkey


class _DeleteAfterDate(_Delete):
    def __init__(self, table, whereFields, whereValues, dateField, dateValue=None, replaceSymbol='%s', notExecMany=False):
        _Delete.__init__(self, table, whereFields, whereValues, replaceSymbol, notExecMany)
        self.dateField = dateField
        self.dateValue = dateValue

    def __str__(self):
        return '[DELETE table=%s, whereFields=%s, whereValues=%s, dateField=%s after dateValue=%s]' % (self.table, self.whereFields, self.whereValues, self.dateField, self.dateValue)

    def prepareStatement(self, context):
        hashkey = self.table + '@' + repr(self.whereFields)+'after'+repr(self.dateField)
        if self.dateValue:
            self.whereValues.append(self.dateValue)
        if hashkey in _Delete.cache:
            sql = _Delete.cache[hashkey]
        else:
            tableName = context.tableName(self.table)
            whereFields = self.whereFields
            whereClause = (' and '.join(['%s = ?'] * len(whereFields)) % tuple(whereFields)) + (' AND %s > ?' % self.dateField)
            whereClause = whereClause.replace('?', self.replaceSymbol)
            sql = 'delete from %s where %s' % (tableName, whereClause)
            _Delete.cache[hashkey] = sql
            _Delete.cacheValues[hashkey] = []
        return sql, hashkey


class _DeleteNotNestedItems(_Delete):
    def __init__(self, table, mainCodeField, mainCodeValue, nestedCodeField, nestedCodeIn):
        pass

    def __str__(self):
        return 'Not execute!'

    def prepareStatement(self, context):
        pass

    def execute(self, context):
        pass


# Увеличение значения поля на заданное значение, если строка существует или
# Добавление новой строки в таблицу
class _IncreaseOrInsertValue(object):

    # increaseField - название поля, которое необходимо увеличить
    # increaseField - заданное значение, на которое необходимо увеличить значение поля
    # values - список пар(название поля - значение), которые необходимо добавить в БД
    # whereFields - список полей, по которым осуществляется поиск поля для увеличения значения
    # whereValues - список значений, по которым осуществляется поиск поля для увеличения значения
    def __init__(self, table, increaseField, increaseValue, values, whereFields, whereValues):
        self.table = table
        self.increaseField = increaseField
        self.increaseValue = increaseValue
        self.values = values
        self.whereFields = whereFields
        self.whereValues = whereValues

    def __str__(self):
        return ('[INCREASE VALUE table=%s, field=%s, value=%s, whereFields=%s, whereValues=%s]'
                              % (self.table, self.increaseField, self.increaseValue, self.whereFields, self.whereValues))

    def getWhereClause(self):
        whereClause = ""
        for i in range(len(self.whereFields)):
            if whereClause:
                whereClause += ' and '
            whereClause += "%s = %s" % (self.whereFields[i], self.whereValues[i])
        return whereClause

    def prepareSelectStatement(self, context):
        tableName = context.tableName(self.table)
        sql = ('select %s from %s where %s' % (self.increaseField, tableName, self.getWhereClause()))
        return sql

    def execute(self, context):
        # Поиск поля для увеличения значения
        sql = self.prepareSelectStatement(context)
        context.cursor.execute(sql)
        result = context.cursor.fetchone()
        if result is not None:
             self.increaseValue += result[0]
        self.values[self.increaseField] = self.increaseValue
        for i in range(len(self.whereFields)):
            self.values[self.whereFields[i]] = self.whereValues[i]
        # Добавление строки или обновление значений
        insertOrUpdate = _Insert(self.table, self.values, onDuplicateKeyUpdate=True)
        insertOrUpdate.execute(context)
