# -*- coding: UTF-8 -*-

import ConfigParser
import atexit
import json
import nes_common.logger as logger
import optparse
import os
import signal
import sys
import tempfile
import threading
import time
import traceback
from artix.common import servicelock
from artix.filequeue.filequeuemanager import FileQueueManager
from bottle import route, run, response, request, Bottle
from httplib import INTERNAL_SERVER_ERROR
from nes_common.basictaskrunner import BasicTaskRunner
from nes_common.mproctaskrunner import MprocTaskRunner
from transaction import Transaction


prefixQueueName = 'artix.loaddict.'
timeShopsReread = 0


def getIntFromConfig(config, nameOfParam, defaultValue, section='QUEUE'):
    return config.getint(section, nameOfParam) if config.has_section(section) and config.has_option(section, nameOfParam) else defaultValue


def getStrFromConfig(config, nameOfParam, defaultValue, section='QUEUE'):
    return config.get(section, nameOfParam) if config.has_section(section) and config.has_option(section, nameOfParam) else defaultValue


def getBooleanFromConfig(config, nameOfParam, defaultValue, section='QUEUE'):
    return config.getboolean(section, nameOfParam) if config.has_section(section) and config.has_option(section, nameOfParam) else defaultValue


def receiveMessages(shopDest, isCash):
    shop = {'shopId': shopDest.split('.')[0]}
    shop['clusterId'] = shopDest.replace(shop['shopId'] + '.', '') if shop['shopId'] != shopDest else ''
    log = logger.getLogger("Shop", shop['shopId'])
    queueName = prefixQueueName + shopDest

    try:
        fileQueueManager = FileQueueManager(queueRoot=Transaction.fileQueueRoot, log=log, location = 'cash' if isCash else 'cs')
        while True:
            transactionId = fileQueueManager.getCompleteTransactionId(queueName)
            if not transactionId or not Transaction(shop, log, queueName, transactionId, isCash).handle():
                break
    except Exception as e:
        log.error('Ошибка: %s' % str(e))
        return False
    return True


class RestReceiver(Bottle):
    def __init__(self, taskRunner, receiveMethod):
        super(RestReceiver, self).__init__()
        self.log = logger.getLogger('Service')
        self.taskRunner = taskRunner
        self.receiveMethod = receiveMethod
        self.route('/nes/queue_reread', method='PUT', callback=self.__queueReread)
        self.route('/nes/queue_load', method='PUT', callback=self.__load)
        self.route('/nes/forgot_bad', method='PUT', callback=self.__forgotBad)

    def daemonize(self, host, port):
        self.log.debug('runing REST service on %s:%s...' % (host, port))
        t = threading.Thread(target=self.run, kwargs={'host': host, 'port': port})
        t.daemon = True
        t.start()

    def __queueReread(self):
        global timeShopsReread
        response.content_type = 'application/json'
        result = {u'success': False}
        try:
            timeShopsReread = 0
            self.log.info(u'Change timeShopsReread =  %s' % timeShopsReread)
            result = {u'success': True}
        except Exception as e:
            self.log.warn(u'Ошибка REST сервиса. %s' % str(e))
            self.log.warn(traceback.format_exc())
            response.status = INTERNAL_SERVER_ERROR
        return json.dumps(result, ensure_ascii=False)

    def __load(self):
        body = json.loads(request.body.read())
        shopId = str(body.get('shopId'))
        response.content_type = 'application/json'
        result = {u'success': False}
        try:
            self.log.info('Put new task for load shop:%s' % shopId)
            self.taskRunner.addAsyncTask(shopId, self.receiveMethod, [shopId, False])
            result = {u'success': True}
        except:
            self.log.warn(traceback.format_exc())
            response.status = INTERNAL_SERVER_ERROR
        return json.dumps(result, ensure_ascii=False)

    def __forgotBad(self):
        body = json.loads(request.body.read(), encoding='utf8')
        shopId = body.get('shopId')
        response.content_type = 'application/json'
        result = {u'success': False}
        try:
            self.log.info('Forgot bad task for shop:%s' % shopId)
            self.taskRunner.forgotBadTasks()
            result = {u'success': True}
        except Exception as e:
            self.log.warn(u'Ошибка REST сервиса. %s' % str(e))
            self.log.warn(traceback.format_exc())
            response.status = INTERNAL_SERVER_ERROR
        return json.dumps(result, ensure_ascii=False)


class QueueReceiver(object):
    def __init__(self, isService=True, isCash=True):
        self.log = logger.getLogger('Service')
        self.isService = isService
        self.isCash = isCash
        self.serviceLock = servicelock.ServiceLock(
                os.path.normpath('%s/locks/lockQueueGetter' % tempfile.gettempdir()), self.log)

        signal.signal(signal.SIGTERM, self.__stop)
        atexit.register(self.__unlock)

        config = ConfigParser.RawConfigParser()
        if not config.read(os.path.join(os.path.dirname(os.path.realpath(__file__)), '../config/nes_load.ini')):
            self.log.error("Configuration file 'nes_load.ini' not found")
            sys.exit(1)

        Transaction.fileQueueRoot = getStrFromConfig(config, 'fileQueueRoot',
                '/linuxcash/cash/data/queues' if self.isCash else '/opt/queues', 'FILEQUEUE')
        sendDictGenerateEventOnCS = getBooleanFromConfig(config, 'sendDictGenerateEventOnCS', True, 'MAIN')
        Transaction.defaultNesGenerateCommandMode = 'increment' if sendDictGenerateEventOnCS else 'none'
        Transaction.defaultStatusServer = getStrFromConfig(config, 'defaultStatusServer', 'localhost:8080', 'CS')

        self.taskRunner = BasicTaskRunner() if self.isCash else MprocTaskRunner()
        self.taskRunner.init(config.getint('MAIN', 'poolProcessSize'))

        self.log.debug(u'Режим сервиса = %s' % self.isService)
        self.log.debug(u'Запуск на кассе = %s' % self.isCash)
        if self.isService:
            self.serviceLock.lock(exitOnLock=True)
            if not self.isCash:
                self.__startRestServiceForCS(config)

    def start(self):
        self.log.info(u'Старт NES на %s %s' % (u'кассе' if self.isCash else u'КС', u'в режиме сервиса' if self.isService else ''))
        shopDests = []
        global timeShopsReread
        while True:
            if self.isCash:
                shopDests = ['cash']
            else:
                if timeShopsReread == 0 or (time.time() - 600) > timeShopsReread:
                    timeShopsReread = time.time()
                    self.log.debug('Получение списка магазинов')
                    tmpShopDests = FileQueueManager(queueRoot=Transaction.fileQueueRoot, log=self.log, location='cs').getShopDests(
                            prefixQueueName)
                    if tmpShopDests:
                        shopDests = tmpShopDests
                    self.log.debug('Список магазинов = %s' % shopDests)

            for shopDest in shopDests:
                self.taskRunner.addAsyncTask(shopDest, receiveMessages, [shopDest, self.isCash])

            if self.isService:
                time.sleep(20)
            else:
                break

    def __stop(self, signum, frame):
        self.log.info(u'Останов очереди')
        if self.taskRunner:
            self.taskRunner.stopAllTasks()
        self.log.info(u'Выход')
        exit(0)

    def __unlock(self):
        self.serviceLock.unlock()

    def __startRestServiceForCS(self, config):
        restServiceHost = getStrFromConfig(config, 'host', '0.0.0.0', 'REST_SERVICE')
        restServicePort = getIntFromConfig(config, 'port', 28081, 'REST_SERVICE')
        self.restReceiver = RestReceiver(self.taskRunner, receiveMessages)
        try:
            self.restReceiver.daemonize(restServiceHost, restServicePort)
        except Exception as e:
            self.log.warn(u'Ошибка запуска REST сервиса: %s' % str(e))
            self.log.warn(traceback.format_exc())


if __name__ == '__main__':
    parser = optparse.OptionParser()
    parser.add_option("-s", "--service", action="store_true", dest="isService", help=u'Старт сервиса', default=False)
    parser.add_option("-t", "--type", dest="type", help=u'Тип сервиса(cash или cs)', default="cash")
    (options, args) = parser.parse_args(sys.argv)
    receiver = QueueReceiver(options.isService, options.type == "cash")
    receiver.start()
