# -*- coding: UTF-8 -*-

import mysql.db_inserter
import tempfile
from _mysql_exceptions import OperationalError, DatabaseError
from artix.common.progress import Progress
from artix.filequeue.filequeuemanager import FileQueueManager
from nes_common import db_request_executor, rest_client


class Transaction(object):
    # специально статические, т.к. инициализируются единожды при старте сервиса
    defaultNesGenerateCommandMode = 'none'
    defaultStatusServer = 'localhost:8080'
    fileQueueRoot = tempfile.gettempdir()

    """ Класс одной полной транзакции загрузки справочника """
    def __init__(self, shop, log, queueName, transactionId, isCash):
        self.shop = shop
        self.log = log
        self.queueName = queueName
        self.transactionId = transactionId
        self.isCash = isCash
        self.fileQueueManager = None
        self.inserter = None
        self.generateList = {}
        self.dictBackOfficeIds = set()
        self.tableNames = set()
        self.countGoodCommands = {'add': 0, 'delete': 0, 'clear': 0, 'update': 0}
        self.countBadCommands = {'add': 0, 'delete': 0, 'clear': 0, 'update': 0}
        self.progress = Progress(log)

    def handle(self):
        """ Обработка транзакции
        Returns:
            bool: True - обработка транзакции с удалением из очереди, False - иначе
        """
        try:
            if self.__prepare() and self.__process():
                self.fileQueueManager.removeTransaction(self.queueName, self.transactionId)
                return True
        except Exception as e:
            self.log.error(str(e))
        finally:
            self.progress.removeProgresses()
        return False

    def __prepare(self):
        """ Подготовка транзакции
        Returns:
            bool: True - успех, False - иначе
        """
        self.log.info('Подготовка к загрузке справочников...')
        self.generateList = {}
        self.dictBackOfficeIds = set()
        self.tableNames = set()
        self.countGoodCommands = {'add': 0, 'delete': 0, 'clear': 0, 'update': 0}
        self.countBadCommands = {'add': 0, 'delete': 0, 'clear': 0, 'update': 0}
        try:
            self.fileQueueManager = FileQueueManager(queueRoot=Transaction.fileQueueRoot, log=self.log, location='cash' if self.isCash else 'cs')
        except Exception as e:
            self.log.error('Ошибка создания менеджера файл-очередей: %s' % str(e))
            return False
        self.transactionProperties = self.fileQueueManager.getTransactionProperties(self.queueName, self.transactionId)
        if not self.transactionProperties:
            self.log.error('Ошибка получения свойств транзакции')
            self.fileQueueManager.removeTransaction(self.queueName, self.transactionId)
            return False
        self.loadDictStatusSender = self.__createLoadDictSender()
        self.progress.createProgress(self.transactionProperties.get('transactionId'), self.transactionProperties.get('objectsCount'))
        return True

    def __process(self):
        """ Обработка сообщений транзакции
        Returns:
            bool: True - завершить обработку транзакции и удалить её, False - завершить обработку транзакции и не удалять её
        """
        messageIter = iter(self.fileQueueManager.getNextMessage(self.queueName, self.transactionId))
        while True:
            try:
                message = next(messageIter, None)
            except:
                self.loadDictStatusSender.sendDictState('Error while getting message',
                        'Ошибка получения сообщения из очереди', rest_client.STATUS_ERROR)
                return True

            # сообщений в очереди больше нет - когда транзакция завершается внезапно без сообщения end_tran
            if not message:
                self.__unexpectedTransactionEnd()
                return True

            if message.get('properties').get('type') == 'comparetable':
                # comparetable - транзакция из одного файла
                self.__compareTable(message)
                return True

            if message.get('properties').get('type') == 'datagram':
                if not self.__datagram(message):
                    return False
            elif message.get('properties').get('control') == 'end_tran':
                return self.__transactionEnd()
            elif message.get('properties').get('control') == 'begin_tran':
                pass
            else:
                self.log.warning('Получили сообщение неизвестного формата')

    def __unexpectedTransactionEnd(self):
        """ Завершение внезапно завершившейся транзакции - отсутствие файла с данными об окончании транзакции """
        self.__generateDict()
        if self.loadDictStatusSender:
            self.loadDictStatusSender.sendDictState('Not found message "end transaction"',
                                                    'Не найдено сообщение "конец транзакции"', rest_client.STATUS_ERROR)
        self.log.error("Не найдено сообщение 'конец транзакции'")
        if self.inserter:
            self.inserter.stop()

    def __compareTable(self, message):
        """ Обработка сообщения с потабличной сверкой """
        request = self.transactionProperties.get('requestCash')
        compareName = self.transactionProperties.get('compareName')
        compareResultServerId = self.transactionProperties.get('compareResultServerId')
        statusServer = self.transactionProperties.get('statusServer')
        cashCode = self.transactionProperties.get('cashCode')
        self.log.info(u"Получено сообщение для сверки данных <%s>. request: %s" % (compareName, request))
        self.log.debug(u"Доп. пораметры: compareResultServerId = %s, statusServer = %s, cashId = %s"
                       % (compareResultServerId, statusServer, cashCode))
        try:
            requestExecutor = db_request_executor.DB_RequestExecutor(self.shop, self.isCash)
            self.log.info("Выполняем запрос в БД")
            executeResult = requestExecutor.execute(request)
            resultServer = message.get('content')
            if resultServer is not None:
                self.log.debug(u"Резальтат выполнения запроса на КС: %s" % resultServer)
            self.log.debug(u"Резальтат выполнения запроса: %s" % (str(executeResult).decode('string-escape').decode('utf8')))
            self.log.info("Отправляем результат запроса на сервер")
            rest_client.sendCompareTableResult(compareResultServerId, self.shop.get('shopId', ''),
                                               cashCode, executeResult, statusServer)
        except Exception as e:
            self.log.error(str(e))

    def __datagram(self, message):
        """ Обработка сообщения с данными справочника
        Returns:
            bool: True - успех, False - иначе
        """
        self.log.info('Добавляем объекты в БД')
        if not self.__createInserter():
            return False
        try:
            tables, goodCount, badCount = self.inserter.add_package(message.get('content'))
            self.tableNames = self.tableNames.union(tables)
            for key in self.countGoodCommands.keys():
                self.countGoodCommands[key] += goodCount[key]
                self.countBadCommands[key] += badCount[key]
        except Exception as e:
            self.log.error('Ошибка при добавлении объектов на загрузку в БД: %s' % e)
            try:
                self.inserter.flagUnBlockSet()
            except Exception as e:
                self.log.error('Ошибка разблокировки БД и закрытия соединения: %s' % e)
            self.loadDictStatusSender.sendDictState(str(e), 'Ошибка: %s' % str(e), rest_client.STATUS_ERROR)
        return True

    def __transactionEnd(self):
        """ Обработка сообщения с данными окончания транзакции
        Returns:
            bool: True - успех, False - иначе
        """
        self.__generateDict()
        if sum(self.countGoodCommands.values()) + sum(self.countBadCommands.values()) == 0:
            self.log.warning('Не получено ни одного сообщения с объектами из справочника')
        self.log.info('Закончили обход объектов в текущей транзакции')
        if self.inserter and not self.inserter.stop():
            self.loadDictStatusSender.sendDictState('Stop failure insert into DB',
                                                    'Ошибка завершения вставки в БД', rest_client.STATUS_ERROR)
            self.log.info('Попробуем выполнить загрузку позже')
            return False
        self.log.info('Транзакция для магазина - %s окончена' % self.shop['shopId'])
        mes_ru, mes_en = self.__dictsCountCommandsToStr()
        self.log.info(mes_ru)
        if sum(self.countBadCommands.values()) > 0:
            self.loadDictStatusSender.sendDictState('Dictionary is not fully loaded. ' + mes_en,
                                                    'Справочник загружен, но не полностью. Были ошибки. ' + mes_ru,
                                                    rest_client.STATUS_READY_WITH_ERROR, self.tableNames)
            # при загрузке в КС для остальных таблиц, у которых была успешная загрузка ставим "хороший" статус
            if not self.loadDictStatusSender.cashCode:
                self.loadDictStatusSender.sendDictState('OK ' + mes_en, 'ОК ' + mes_ru, rest_client.STATUS_READY)
        else:
            self.loadDictStatusSender.sendDictState('OK ' + mes_en, 'ОК ' + mes_ru, rest_client.STATUS_READY)
        return True

    def __isSendLoadDictStatus(self):
        """ Получение флага отправки статуса загрузки на сервер
        Returns:
            bool: True - отправлять статус на сервер, False - иначе
        """
        isSendLoadDictStatus = True
        nesGenerateCommandMode = self.transactionProperties.get('nesGenerateCommandMode', self.defaultNesGenerateCommandMode)
        if nesGenerateCommandMode != 'none':
            unloadElementName = self.transactionProperties.get('unloadElementName', None)
            if unloadElementName:
                nesGenerateCommandCashesList = [x.strip() for x in
                        self.transactionProperties.get('nesGenerateCommandCashesList', '').split(',')]
                # если пришла просто команда на генерацию справочника без самого справочника, то не будем отправлять статус загрузки в БД
                isSendLoadDictStatus = False
                if unloadElementName in self.generateList.get(nesGenerateCommandMode, {}):
                    self.generateList[nesGenerateCommandMode][unloadElementName] += nesGenerateCommandCashesList
                else:
                    if nesGenerateCommandMode not in self.generateList:
                        self.generateList[nesGenerateCommandMode] = {}
                    self.generateList[nesGenerateCommandMode][unloadElementName] = nesGenerateCommandCashesList
            else:
                # если передали статус генерации на КС и режим генерации не установлен в none и не передали unloadElementName,
                # то для этого справочника будем искать unloadElement-ы для запуска генерации
                dictCSStatusId = self.transactionProperties.get('dictBackOfficeId')
                if dictCSStatusId:
                    self.dictBackOfficeIds.add(dictCSStatusId)
        return isSendLoadDictStatus

    def __createLoadDictSender(self):
        """ Создание отправителя статуса загрузки справочников на сервер
        Returns:
            LoadDictStatusSender: объект отправки статуса загрузки справочников на сервер
        """
        dictCSStatusId = self.transactionProperties.get('dictBackOfficeId')
        isSendLoadDictStatus = self.__isSendLoadDictStatus()
        dictCashStatusId = self.transactionProperties.get('dictCashLoadStateId')
        cashCode = self.transactionProperties.get('cashCode')
        statusServer = self.transactionProperties.get('statusServer')
        version = self.transactionProperties.get('version')
        return rest_client.LoadDictStatusSender(self.shop['shopId'], dictCSLoadStatusId=dictCSStatusId,
                                                dictCashLoadStatusId=dictCashStatusId, statusServer=statusServer,
                                                cashCode=cashCode, version=version,
                                                isSendLoadDictStatus=isSendLoadDictStatus,
                                                defaultStatusServer=self.defaultStatusServer)

    def __createInserter(self):
        """ Создание объекта для вставки данных в БД
        Returns:
            bool: True - успешное создание, False - иначе
        """
        if self.inserter:
            return True
        try:
            self.inserter = mysql.db_inserter.DB_Inserter(self.shop, self.loadDictStatusSender)
        except (DatabaseError, OperationalError) as e:
            self.log.error('Ошибка подключения к БД. %s' % str(e))
            self.log.info('Попробуем выполнить загрузку позже')
            return False
        return self.inserter.start(self.transactionProperties.get('transactionId'))

    def __dictsCountCommandsToStr(self):
        """ Получение текстовых сообщений о количестве успешных и неуспешных команд для каждого вида команды
        Returns:
            (str, str): сообщения
        """
        clear_ru, clear_en = self.__dictsCountCommandToStr('clear', 'очистки')
        add_ru, add_en = self.__dictsCountCommandToStr('add', 'добавления')
        delete_ru, delete_en = self.__dictsCountCommandToStr('delete', 'удаления')
        update_ru, update_en = self.__dictsCountCommandToStr('update', 'обновления')
        return clear_ru + add_ru + delete_ru + update_ru, clear_en + add_en + delete_en + update_en

    def __dictsCountCommandToStr(self, command, ruName):
        if self.countGoodCommands[command] + self.countBadCommands[command] > 0:
            s_ru = 'Команды %s: успешно - %s, ошибок - %s; ' % (ruName, self.countGoodCommands[command], self.countBadCommands[command])
            s_en = '%s commans: ok - %s, error - %s; ' % (command, self.countGoodCommands[command], self.countBadCommands[command])
            return s_ru, s_en
        return '', ''

    def __generateDict(self):
        if self.dictBackOfficeIds or self.generateList:
            self.log.info('Запускаем генерацию справоников на КС')
            rest_client.generateDict(self.shop['shopId'], self.dictBackOfficeIds,
                                     self.transactionProperties.get('nesGenerateCommandOwner', ''),
                                     self.generateList, self.defaultStatusServer)
